/***************************************************************************
 $RCSfile: adduser.c,v $
 -------------------
 cvs         : $Id: adduser.c,v 1.2 2005/06/29 03:44:58 aquamaniac Exp $
 begin       : Tue May 03 2005
 copyright   : (C) 2005 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "globals.h"

#include <gwenhywfar/text.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>


AH_MEDIUM *_getMedium(AH_HBCI *hbci,
		      const char *mtype,
		      const char *mname) {
  int rv;
  AH_MEDIUMDEVICE dt;
  GWEN_BUFFER *bufMTypeName;
  GWEN_BUFFER *bufMName;
  AH_MEDIUM *m=0;

  if (strcasecmp(mtype, "card")==0)
    dt=AH_MediumDeviceChipcard;
  else if (strcasecmp(mtype, "file")==0)
    dt=AH_MediumDeviceFile;
  else {
    DBG_ERROR(0, "Unknown medium device type \"%s\"", mtype);
    return 0;
  }
  bufMTypeName=GWEN_Buffer_new(0, 256, 0, 1);
  bufMName=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AH_HBCI_CheckMedium(hbci, dt, bufMTypeName, bufMName);
  if (rv==0) {
    DBG_NOTICE(0, "Medium type: %s", GWEN_Buffer_GetStart(bufMTypeName));
    DBG_NOTICE(0, "Medium name: %s", GWEN_Buffer_GetStart(bufMName));
    m=AH_HBCI_MediumFactory(hbci,
			    GWEN_Buffer_GetStart(bufMTypeName),
			    GWEN_Buffer_GetStart(bufMName));
    if (!m) {
      DBG_ERROR(0, "Could not create medium of supported type \"%s\"",
		GWEN_Buffer_GetStart(bufMTypeName));
    }
  }
  else {
    DBG_ERROR(0, "Medium not supported");
  }
  GWEN_Buffer_free(bufMName);
  GWEN_Buffer_free(bufMTypeName);
  return m;
}


int getBankUrl(AB_BANKING *ab,
	       AH_MEDIUMTYPE mt,
	       const char *bankId,
	       GWEN_BUFFER *bufServer) {
  AB_BANKINFO *bi;
  
  bi=AB_Banking_GetBankInfo(ab, "de", 0, bankId);
  if (bi) {
    AB_BANKINFO_SERVICE_LIST *l;
    AB_BANKINFO_SERVICE *sv;
  
    l=AB_BankInfo_GetServices(bi);
    assert(l);
    sv=AB_BankInfoService_List_First(l);
    while(sv) {
      const char *st;
  
      st=AB_BankInfoService_GetType(sv);
      if (st && *st && strcasecmp(st, "hbci")==0) {
	const char *svm;
  
	svm=AB_BankInfoService_GetMode(sv);
	if (svm && *svm) {
	  if (!
	      ((strcasecmp(svm, "pintan")==0) ^
	       (mt==AH_MediumTypePINTAN))){
	    const char *addr;
  
	    addr=AB_BankInfoService_GetAddress(sv);
	    if (addr && *addr) {
	      GWEN_Buffer_Reset(bufServer);
	      GWEN_Buffer_AppendString(bufServer, addr);
              return 0;
	    }
	  }
	}
      }
      sv=AB_BankInfoService_List_Next(sv);
    }
    AB_BankInfo_free(bi);
  } /* if bank info */

  return -1;
}



int addUser(AB_BANKING *ab,
            GWEN_DB_NODE *dbArgs,
            int argc,
            char **argv) {
  GWEN_DB_NODE *db;
  AB_PROVIDER *pro;
  AH_HBCI *hbci;
  int rv;
  const char *s;
  const char *bankId;
  const char *userId;
  const char *customerId;
  AH_MEDIUM *medium=0;
  const char *mediumName;
  const char *server;
  const GWEN_ARGS args[]={
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "bankId",                     /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "b",                          /* short option */
    "bank",                       /* long option */
    "Specify the bank code",      /* short description */
    "Specify the bank code"       /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "userId",                     /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "u",                          /* short option */
    "user",                       /* long option */
    "Specify the user id",        /* short description */
    "Specify the user id"         /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "customerId",                 /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "c",                          /* short option */
    "customer",                   /* long option */
    "Specify the customer id",    /* short description */
    "Specify the customer id"     /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "mediumType",                 /* name */
    1,                            /* minnum */
    1,                            /* maxnum */
    0,                            /* short option */
    "mediumtype",                 /* long option */
    "Specify the medium type (file or card)",      /* short description */
    "Specify the medium type (file or card)"       /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "mediumName",                 /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "m",                          /* short option */
    "mediumname",                 /* long option */
    "Specify the medium name",    /* short description */
    "Specify the medium name"     /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeChar,            /* type */
    "serverAddr",                 /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "s",                          /* short option */
    "server",                     /* long option */
    "Specify the server URL",     /* short description */
    "Specify the server URL"      /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HAS_ARGUMENT, /* flags */
    GWEN_ArgsTypeInt,             /* type */
    "context",                    /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    0,                            /* short option */
    "context",                    /* long option */
    "Select a context on the medium", /* short description */
    "Select a context on the medium"  /* long description */
  },
  {
    0, /* flags */
    GWEN_ArgsTypeInt,             /* type */
    "import",                     /* name */
    0,                            /* minnum */
    1,                            /* maxnum */
    "i",                          /* short option */
    "import",                     /* long option */
    "Import an existing medium",  /* short description */
    "Import an existing medium"      /* long description */
  },
  {
    GWEN_ARGS_FLAGS_HELP | GWEN_ARGS_FLAGS_LAST, /* flags */
    GWEN_ArgsTypeInt,             /* type */
    "help",                       /* name */
    0,                            /* minnum */
    0,                            /* maxnum */
    "h",                          /* short option */
    "help",                       /* long option */
    "Show this help screen",      /* short description */
    "Show this help screen"       /* long description */
  }
  };

  db=GWEN_DB_GetGroup(dbArgs, GWEN_DB_FLAGS_DEFAULT, "local");
  rv=GWEN_Args_Check(argc, argv, 1,
                     0 /*GWEN_ARGS_MODE_ALLOW_FREEPARAM*/,
                     args,
                     db);
  if (rv==GWEN_ARGS_RESULT_ERROR) {
    fprintf(stderr, "ERROR: Could not parse arguments\n");
    return 1;
  }
  else if (rv==GWEN_ARGS_RESULT_HELP) {
    GWEN_BUFFER *ubuf;

    ubuf=GWEN_Buffer_new(0, 1024, 0, 1);
    if (GWEN_Args_Usage(args, ubuf, GWEN_ArgsOutTypeTXT)) {
      fprintf(stderr, "ERROR: Could not create help string\n");
      return 1;
    }
    fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(ubuf));
    GWEN_Buffer_free(ubuf);
    return 0;
  }

  rv=AB_Banking_Init(ab);
  if (rv) {
    DBG_ERROR(0, "Error on init (%d)", rv);
    return 2;
  }

  pro=AB_Banking_GetProvider(ab, "aqhbci");
  assert(pro);
  hbci=AH_Provider_GetHbci(pro);
  assert(hbci);

  bankId=GWEN_DB_GetCharValue(db, "bankId", 0, 0);
  userId=GWEN_DB_GetCharValue(db, "userId", 0, 0);
  customerId=GWEN_DB_GetCharValue(db, "customerId", 0, 0);
  mediumName=GWEN_DB_GetCharValue(db, "mediumName", 0, 0);
  server=GWEN_DB_GetCharValue(db, "serverAddr", 0, 0);
  s=GWEN_DB_GetCharValue(db, "mediumType", 0, 0);
  assert(s && *s);
  if (GWEN_DB_VariableExists(db, "import") || strcasecmp(s, "card")==0) {
    /* import existing medium */
    medium=_getMedium(hbci, s, mediumName);
  }
  else {
    medium=AH_HBCI_SelectMedium(hbci, s, mediumName);
  }
  if (!medium) {
    DBG_ERROR(0, "Medium type \"%s\" not supported", s);
    return 3;
  }

  if (GWEN_DB_VariableExists(db, "import")) {
    int idx;
    int rv;

    rv=AH_Medium_Mount(medium);
    if (rv) {
      DBG_ERROR(0, "Error mounting medium (%d)", rv);
      return 3;
    }
    idx=GWEN_DB_GetIntValue(db, "context", 0, -1);
    if (idx<0) {
      idx=AH_Medium_FindContext(medium,
				280,
				bankId?bankId:"*",
				userId?userId:"*");
      if (idx<0) {
	DBG_ERROR(0, "No context given and no matching found");
	return 3;
      }
    }
    if (idx>=0) {
      int country;
      GWEN_BUFFER *bufBankId;
      GWEN_BUFFER *bufUserId;
      GWEN_BUFFER *bufServer;
      int port;
      int rv;
      AH_BANK *bank;
      AH_USER *user;
      AH_CUSTOMER *customer;
      AH_BPD_ADDR *ba;
      const char *lbankId;
      const char *luserId;
      const char *lcustomerId;
      const char *lserverAddr;

      bufBankId=GWEN_Buffer_new(0, 32, 0, 1);
      bufUserId=GWEN_Buffer_new(0, 32, 0, 1);
      bufServer=GWEN_Buffer_new(0, 32, 0, 1);
      rv=AH_Medium_ReadContext(medium,
			       idx,
			       &country,
			       bufBankId,
			       bufUserId,
			       bufServer,
			       &port);
      if (rv) {
	DBG_ERROR(0, "Error reading context %d from medium", idx);
        return 3;
      }

      lbankId=bankId?bankId:GWEN_Buffer_GetStart(bufBankId);
      bank=AH_HBCI_FindBank(hbci, 280, lbankId);
      if (!bank) {
	bank=AH_Bank_new(hbci, 280, lbankId);
	assert(bank);
	AH_HBCI_AddBank(hbci, bank);
      }
      luserId=userId?userId:GWEN_Buffer_GetStart(bufUserId);
      user=AH_Bank_FindUser(bank, luserId);
      if (user) {
	DBG_ERROR(0, "User %s already exists", luserId);
	GWEN_Buffer_free(bufServer);
	GWEN_Buffer_free(bufUserId);
	GWEN_Buffer_free(bufBankId);
        return 3;
      }
      user=AH_User_new(bank, luserId, medium);
      assert(user);
      AH_Bank_AddUser(bank, user);
      AH_User_SetContextIdx(user, idx);

      lcustomerId=customerId?customerId:luserId;
      customer=AH_User_FindCustomer(user, lcustomerId);
      if (customer) {
	DBG_ERROR(0, "Customer %s already exists", lcustomerId);
	GWEN_Buffer_free(bufServer);
	GWEN_Buffer_free(bufUserId);
	GWEN_Buffer_free(bufBankId);
	return 3;
      }

      customer=AH_Customer_new(user, lcustomerId);
      assert(customer);
      AH_User_AddCustomer(user, customer);

      /* try to get server address from database if still unknown */
      if (!server && GWEN_Buffer_GetUsedBytes(bufServer)==0) {
	if (getBankUrl(ab,
                       AH_Medium_GetMediumType(medium),
		       lbankId,
		       bufServer)) {
	  DBG_INFO(0, "Could not find server address for \"%s\"",
		   lbankId);
	}
      }
      lserverAddr=server?server:GWEN_Buffer_GetStart(bufServer);
      if (!lserverAddr || !*lserverAddr) {
	DBG_ERROR(0, "No address given and none available in internal db");
	GWEN_Buffer_free(bufServer);
	GWEN_Buffer_free(bufUserId);
	GWEN_Buffer_free(bufBankId);
	return 3;
      }

      /* set address */
      ba=AH_BpdAddr_new();
      AH_BpdAddr_SetAddr(ba, lserverAddr);
      if (AH_Medium_GetMediumType(medium)==AH_MediumTypePINTAN) {
	AH_BpdAddr_SetType(ba, AH_BPD_AddrTypeSSL);
        AH_BpdAddr_SetSuffix(ba, "443");
      }
      else {
	AH_BpdAddr_SetType(ba, AH_BPD_AddrTypeTCP);
        AH_BpdAddr_SetSuffix(ba, "3000");
      }
      AH_User_SetAddress(user, ba);
      AH_BpdAddr_free(ba);

      GWEN_Buffer_free(bufServer);
      GWEN_Buffer_free(bufUserId);
      GWEN_Buffer_free(bufBankId);

      if (AH_Medium_GetMediumType(medium)==AH_MediumTypeDDV)
	AH_User_SetStatus(user, AH_UserStatusEnabled);
    }
    else {
      DBG_ERROR(0, "Invalid context %d", idx);
      return 3;
    }
  } /* if import mode */
  else {
    GWEN_BUFFER *bufServer=0;
    const char *lserverAddr;
    const char *lcustomerId;
    int ctx;
    AH_BANK *bank;
    AH_USER *user;
    AH_CUSTOMER *customer;
    AH_BPD_ADDR *ba;

    /* create context */
    if (!userId || !bankId) {
      DBG_ERROR(0,
		"User ID or bank code missing");
      return 1;
    }
    if (!server) {
      bufServer=GWEN_Buffer_new(0, 64, 0, 1);
      if (getBankUrl(ab,
		     AH_Medium_GetMediumType(medium),
		     bankId,
		     bufServer)) {
	DBG_ERROR(0,
		  "No address given for \"%s\" "
		  "and none available in internal db",
		  bankId);
	GWEN_Buffer_free(bufServer);
	return 3;
      }
    }
    lserverAddr=server?server:GWEN_Buffer_GetStart(bufServer);

    /* create and/or setup bank */
    bank=AH_HBCI_FindBank(hbci, 280, bankId);
    if (!bank) {
      bank=AH_Bank_new(hbci, 280, bankId);
      assert(bank);
      AH_HBCI_AddBank(hbci, bank);
    }
    user=AH_Bank_FindUser(bank, userId);
    if (user) {
      DBG_ERROR(0, "User %s already exists", userId);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    /* create and setup user */
    user=AH_User_new(bank, userId, medium);
    assert(user);
    /* set address */
    ba=AH_BpdAddr_new();
    AH_BpdAddr_SetAddr(ba, lserverAddr);
    if (AH_Medium_GetMediumType(medium)==AH_MediumTypePINTAN) {
      AH_BpdAddr_SetType(ba, AH_BPD_AddrTypeSSL);
      AH_BpdAddr_SetSuffix(ba, "443");
    }
    else {
      AH_BpdAddr_SetType(ba, AH_BPD_AddrTypeTCP);
      AH_BpdAddr_SetSuffix(ba, "3000");
    }
    AH_User_SetAddress(user, ba);
    AH_BpdAddr_free(ba);

    AH_Bank_AddUser(bank, user);

    /* create customer */
    lcustomerId=customerId?customerId:userId;
    customer=AH_User_FindCustomer(user, lcustomerId);
    if (customer) {
      DBG_ERROR(0, "Customer %s already exists", lcustomerId);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    customer=AH_Customer_new(user, lcustomerId);
    assert(customer);
    AH_User_AddCustomer(user, customer);

    /* create medium */
    rv=AH_Medium_Create(medium);
    if (rv) {
      DBG_ERROR(0, "Error creating medium (%d)", rv);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    /* mount medium */
    rv=AH_Medium_Mount(medium);
    if (rv) {
      DBG_ERROR(0, "Error mounting medium (%d)", rv);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    ctx=AH_Medium_CreateContext(medium, 280, bankId, userId);
    if (ctx<0) {
      DBG_ERROR(0,
		"Could not create user \"%s:%s\" on medium",
		bankId, userId);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    AH_User_SetContextIdx(user, ctx);

    rv=AH_Medium_Unmount(medium, 1);
    if (rv) {
      DBG_ERROR(0, "Error unmounting medium (%d)", rv);
      GWEN_Buffer_free(bufServer);
      return 3;
    }

    GWEN_Buffer_free(bufServer);
  } /* if not in import mode */

  rv=AB_Banking_Fini(ab);
  if (rv) {
    fprintf(stderr, "ERROR: Error on deinit (%d)\n", rv);
    return 5;
  }

  return 0;
}




